// Copyright 2004 Adam Megacz, see the COPYING file for licensing [GPL]
package org.ibex;

import org.ibex.js.*;
import org.ibex.util.*;

/** 
 *  A Surface, as described in the Ibex Reference.
 *
 *  Platform subclasses should include an inner class subclass of
 *  Surface to return from the Platform._createSurface() method
 */
public abstract class Surface extends PixelBuffer implements Scheduler.Task {

    // Static Data ////////////////////////////////////////////////////////////////////////////////

    private static Boolean T = Boolean.TRUE;
    private static Boolean F = Boolean.FALSE;

    /** all instances of Surface which need to be refreshed by the Scheduler */
    public static Vec allSurfaces = new Vec();
    
    /** When set to true, render() should abort as soon as possible and restart the rendering process */
    volatile boolean abort = false;

    // these three variables are used to ensure that user resizes trump programmatic resizes
    volatile boolean syncRootBoxToSurface = false;
    volatile int pendingWidth = 0;
    volatile int pendingHeight = 0;

    public static boolean alt = false;          ///< true iff the alt button is pressed down
    public static boolean control = false;      ///< true iff the control button is pressed down
    public static boolean shift = false;        ///< true iff the shift button is pressed down
    public static boolean button1 = false;      ///< true iff button 1 is depressed
    public static boolean button2 = false;      ///< true iff button 2 is depressed
    public static boolean button3 = false;      ///< true iff button 3 is depressed


    // Instance Data ///////////////////////////////////////////////////////////////////////

    public Box root;                                   ///< The Box at the root of this surface
    public String cursor = "default";                  ///< The active cursor to switch to when syncCursor() is called
    public int mousex;                                 ///< x position of the mouse
    public int mousey;                                 ///< y position of the mouse
    public int _mousex;                                ///< x position of the mouse FIXME
    public int _mousey;                                ///< y position of the mouse FIXME
    public int newmousex = -1;                         ///< x position of the mouse, in real time; this lets us collapse Move's
    public int newmousey = -1;                         ///< y position of the mouse, in real time; this lets us collapse Move's
    public boolean minimized = false;                  ///< True iff this surface is minimized, in real time
    public boolean maximized = false;                  ///< True iff this surface is maximized, in real time
    DirtyList dirtyRegions = new DirtyList();          ///< Dirty regions on the surface

    // Used For Simulating Clicks and DoubleClicks /////////////////////////////////////////////////

    int last_press_x = Integer.MAX_VALUE;      ///< the x-position of the mouse the last time a Press message was enqueued
    int last_press_y = Integer.MAX_VALUE;      ///< the y-position of the mouse the last time a Press message was enqueued
    static int lastClickButton = 0;            ///< the last button to recieve a Click message; used for simulating DoubleClick's
    static long lastClickTime = 0;             ///< the last time a Click message was processed; used for simulating DoubleClick's
    
    
    // Methods to be overridden by subclasses ///////////////////////////////////////////////////////

    public abstract void toBack();                     ///< should push surface to the back of the stacking order
    public abstract void toFront();                    ///< should pull surface to the front of the stacking order
    public abstract void syncCursor();                 ///< set the actual cursor to this.cursor if they do not match
    public abstract void setInvisible(boolean b);      ///< If <tt>b</tt>, make window invisible; otherwise, make it non-invisible.
    protected abstract void _setMaximized(boolean b);  ///< If <tt>b</tt>, maximize the surface; otherwise, un-maximize it.
    protected abstract void _setMinimized(boolean b);  ///< If <tt>b</tt>, minimize the surface; otherwise, un-minimize it.
    public abstract void setLocation();                ///< Set the surface's x/y position to that of the root box
    protected abstract void _setSize(int w, int h);    ///< set the actual size of the surface
    public abstract void setTitleBarText(String s);    ///< Sets the surface's title bar text, if applicable
    public abstract void setIcon(Picture i);           ///< Sets the surface's title bar text, if applicable
    public abstract void _dispose();                   ///< Destroy the surface
    public void setMinimumSize(int minx, int miny, boolean resizable) { }
    protected void setSize(int w, int h) { _setSize(w, h); }


    // Helper methods for subclasses ////////////////////////////////////////////////////////////

    protected final void Press(final int button) {
        last_press_x = mousex;
        last_press_y = mousey;

        if (button == 1) button1 = true;
        else if (button == 2) button2 = true;
        else if (button == 3) button3 = true;

        if (button == 1) new Message("_Press1", T, root);
        else if (button == 2) new Message("_Press2", T, root);
        else if (button == 3) {
            Scheduler.add(new Scheduler.Task() { public void perform() throws JSExn {
                Platform.clipboardReadEnabled = true;
                try {
                    root.putAndTriggerTraps("_Press3", T);
                } finally {
                    Platform.clipboardReadEnabled = false;
                }
            }});
        }
    }

    protected final void Release(int button) {
        if (button == 1) button1 = false;
        else if (button == 2) button2 = false;
        else if (button == 3) button3 = false;

        if (button == 1) new Message("_Release1", T, root);
        else if (button == 2) new Message("_Release2", T, root);
        else if (button == 3) new Message("_Release3", T, root);

        if (Platform.needsAutoClick() && Math.abs(last_press_x - mousex) < 5 && Math.abs(last_press_y - mousey) < 5) Click(button);
        last_press_x = Integer.MAX_VALUE;
        last_press_y = Integer.MAX_VALUE;
    }

    protected final void Click(int button) {
        if (button == 1) new Message("_Click1", T, root);
        else if (button == 2) new Message("_Click2", T, root);
        else if (button == 3) new Message("_Click3", T, root);
        if (Platform.needsAutoDoubleClick()) {
            long now = System.currentTimeMillis();
            if (lastClickButton == button && now - lastClickTime < 350) DoubleClick(button);
            lastClickButton = button;
            lastClickTime = now;
        }
    }

    /** we enqueue ourselves in the Scheduler when we have a Move message to deal with */
    private Scheduler.Task mover = new Scheduler.Task() {
            public void perform() {
                if (mousex == newmousex && mousey == newmousey) return;
                int oldmousex = mousex;     mousex = newmousex;
                int oldmousey = mousey;     mousey = newmousey;
                String oldcursor = cursor;  cursor = "default";
                // FIXME: Root (ONLY) gets motion events outside itself (if trapped)
                if (oldmousex != mousex || oldmousey != mousey)
                    root.putAndTriggerTrapsAndCatchExceptions("_Move", T);
                if (!cursor.equals(oldcursor)) syncCursor();
            } };

    /**
     *  Notify Ibex that the mouse has moved. If the mouse leaves the
     *  surface, but the host windowing system does not provide its new
     *  position (for example, a Java MouseListener.mouseExited()
     *  message), the subclass should use (-1,-1).
     */
    protected final void Move(final int newmousex, final int newmousey) {
        this.newmousex = newmousex;
        this.newmousey = newmousey;
        Scheduler.add(mover);
    }

    protected final void HScroll(int pixels) { new Message("_HScroll", new Integer(pixels), root); }
    protected final void VScroll(int pixels) { new Message("_VScroll", new Integer(pixels), root); }
    protected final void HScroll(float lines) { new Message("_HScroll", new Float(lines), root); }
    protected final void VScroll(float lines) { new Message("_VScroll", new Float(lines), root); }

    /** subclasses should invoke this method when the user resizes the window */
    protected final void SizeChange(final int width, final int height) {
        if (pendingWidth == width && pendingHeight == height) return;
        pendingWidth = width;
        pendingHeight = height;
        syncRootBoxToSurface = true;
        abort = true;
        Scheduler.renderAll();
    }

    // FEATURE: can we avoid creating objects here?
    protected final void PosChange(final int x, final int y) {
        Scheduler.add(new Scheduler.Task() { public void perform() throws JSExn {
            root.x = x;
            root.y = y;
            root.putAndTriggerTrapsAndCatchExceptions("PosChange", T);
        }});
    }

    private final String[] doubleClick = new String[] { null, "_DoubleClick1", "_DoubleClick2", "_DoubleClick3" };
    protected final void DoubleClick(int button) { new Message(doubleClick[button], T, root); }
    protected final void KeyPressed(String key) { new Message("_KeyPressed", key, root); }
    protected final void KeyReleased(String key) { new Message("_KeyReleased", key, root); }
    protected final void Close() { new Message("Close", T, root); }
    protected final void Minimized(boolean b) { minimized = b; new Message("Minimized", b ? T : F, root); }
    protected final void Maximized(boolean b) { maximized = b; new Message("Maximized", b ? T : F, root); }
    protected final void Focused(boolean b) { new Message("Focused", b ? T : F, root); }

    private boolean scheduled = false;
    public void Refresh() { if (!scheduled) Scheduler.add(this); scheduled = true; }
    public void perform() { scheduled = false; Scheduler.renderAll(); }

    public final void setMaximized(boolean b) { if (b != maximized) _setMaximized(maximized = b); }
    public final void setMinimized(boolean b) { if (b != minimized) _setMinimized(minimized = b); }


    // Other Methods ///////////////////////////////////////////////////////////////////////////////

    /** Indicates that the Surface is no longer needed */
    public final void dispose(boolean quitIfAllSurfacesGone) {
        if (Log.on) Log.info(this, "disposing " + this);
        allSurfaces.removeElement(this);
        _dispose();
        if (allSurfaces.size() == 0) {
            if (Log.on) Log.info(this, "exiting because last surface was destroyed");
            System.exit(0);
        }
    }

    public void dirty(int x, int y, int w, int h) {
        dirtyRegions.dirty(x, y, w, h);
        Refresh();
    }

    public static Surface fromBox(Box b) {
        // FIXME use a hash table here
        for(int i=0; i<allSurfaces.size(); i++) {
            Surface s = (Surface)allSurfaces.elementAt(i);
            if (s.root == b) return s;
        }
        return null;
    }

    public Surface(Box root) {
        this.root = root;
        root.setMaxWidth(JS.N(Math.min(Platform.getScreenWidth(), root.maxwidth)));
        root.setMaxHeight(JS.N(Math.min(Platform.getScreenHeight(), root.maxheight)));
        Surface old = fromBox(root);
        if (old != null) old.dispose(false);
        else root.removeSelf();
        Refresh();
    }

    private static VectorGraphics.Affine identity = VectorGraphics.Affine.identity();

    /** runs the prerender() and render() pipelines in the root Box to regenerate the backbuffer, then blits it to the screen */
    public synchronized void render() {
        scheduled = false;
        // make sure the root is properly sized
        do {
            abort = false;
            root.pack();
            if (syncRootBoxToSurface) {
                root.setMaxWidth(JS.N(pendingWidth));
                root.setMaxHeight(JS.N(pendingHeight));
                syncRootBoxToSurface = false;
            }
            if (root.maxwidth != root.width || root.maxheight != root.height) {
                // dirty the place where the scar used to be and where it is now
                dirty(0, root.height - Main.scarImage.height, Main.scarImage.width, Main.scarImage.height);
                dirty(0, root.maxheight - Main.scarImage.height, Main.scarImage.width, Main.scarImage.height);
            }
            root.resize(root.x, root.y, root.maxwidth, root.maxheight);
            root.place();
            setSize(root.width, root.height);
            /*String oldcursor = cursor;
            cursor = "default";
            root.putAndTriggerTrapsAndCatchExceptions("_Move", JS.T);
            if (!cursor.equals(oldcursor)) syncCursor();*/
        } while(abort);

        int[][] dirt = dirtyRegions.flush();
        for(int i = 0; dirt != null && i < dirt.length; i++) {
            if (dirt[i] == null) continue;
            int x = dirt[i][0], y = dirt[i][1], w = dirt[i][2], h = dirt[i][3];
            if (x < 0) x = 0;
            if (y < 0) y = 0;
            if (x+w > root.width) w = root.width - x;
            if (y+h > root.height) h = root.height - y;
            if (w <= 0 || h <= 0) continue;

            root.render(0, 0, x, y, x + w, y + h, this, identity);
            drawPicture(Main.scarImage, 0, root.height - Main.scarImage.height, x, y, x+w, y+h);
            
            if (abort) {
                // x,y,w,h is only partially reconstructed, so we must be careful not to re-blit it
                dirtyRegions.dirty(x, y, w, h);
                // put back all the dirty regions we haven't yet processed (including the current one)
                for(int j=i; j<dirt.length; j++)
                    if (dirt[j] != null)
                        dirtyRegions.dirty(dirt[j][0], dirt[j][1], dirt[j][2], dirt[j][3]);
                return;
            }
        }
    }

    // FEATURE: reinstate recycler
    public class Message implements Scheduler.Task {
        
        private Box boxContainingMouse;
        private Object value;
        public String name;
        
        Message(String name, Object value, Box boxContainingMouse) {
            this.boxContainingMouse = boxContainingMouse;
            this.name = name;
            this.value = value;
            Scheduler.add(this);
        }
        
        public void perform() {
            if (name.equals("_KeyPressed")) {
                String value = (String)this.value;
                if (value.toLowerCase().endsWith("shift")) shift = true;     else if (shift) value = value.toUpperCase();
                if (value.toLowerCase().equals("alt")) alt = true;           else if (alt) value = "A-" + value;
                if (value.toLowerCase().endsWith("control")) control = true; else if (control) value = "C-" + value;
                if (value.equals("C-v") || value.equals("A-v")) Platform.clipboardReadEnabled = true;
                this.value = value;
            } else if (name.equals("_KeyReleased")) {
                String value = (String)this.value;
                if (value.toLowerCase().equals("alt")) alt = false;
                else if (value.toLowerCase().equals("control")) control = false;
                else if (value.toLowerCase().equals("shift")) shift = false;
                this.value = value;
            } else if (name.equals("_HScroll") || name.equals("_VScroll")) {
                // FIXME: technically points != pixels
                if (value instanceof Integer)
                    value = new Float(((Integer)value).intValue() * root.fontSize());
            }
            try {
                boxContainingMouse.putAndTriggerTrapsAndCatchExceptions(name, value);
            } finally {
                Platform.clipboardReadEnabled = false;
            }
        }
        public String toString() { return "Message [name=" + name + ", value=" + value + "]"; }
    }


    // Default PixelBuffer implementation /////////////////////////////////////////////////////////

    public static abstract class DoubleBufferedSurface extends Surface {

        public DoubleBufferedSurface(Box root) { super(root); }
        PixelBuffer backbuffer = Platform.createPixelBuffer(Platform.getScreenWidth(), Platform.getScreenHeight(), this);
        DirtyList screenDirtyRegions = new DirtyList();

        public void drawPicture(Picture source, int dx, int dy, int cx1, int cy1, int cx2, int cy2) {
            screenDirtyRegions.dirty(cx1, cy1, cx2 - cx1, cy2 - cy1);
            backbuffer.drawPicture(source, dx, dy, cx1, cy1, cx2, cy2);
        }

        public void drawGlyph(Font.Glyph source, int dx, int dy, int cx1, int cy1, int cx2, int cy2, int argb) {
            screenDirtyRegions.dirty(cx1, cy1, cx2 - cx1, cy2 - cy1);
            backbuffer.drawGlyph(source, dx, dy, cx1, cy1, cx2, cy2, argb);
        }

        public void fillTrapezoid(int x1, int x2, int y1, int x3, int x4, int y2, int color) {
            screenDirtyRegions.dirty(Math.min(x1, x3), y1, Math.max(x2, x4) - Math.min(x1, x3), y2 - y1);
            backbuffer.fillTrapezoid(x1, x2, y1, x3, x4, y2, color);
        }

        public void render() {
            super.render();
            if (abort) return;
            int[][] dirt = screenDirtyRegions.flush();
            for(int i = 0; dirt != null && i < dirt.length; i++) {
                if (dirt[i] == null) continue;
                int x = dirt[i][0];
                int y = dirt[i][1];
                int w = dirt[i][2];
                int h = dirt[i][3];
                if (x < 0) x = 0;
                if (y < 0) y = 0;
                if (x+w > root.width) w = root.width - x;
                if (y+h > root.height) h = root.height - y;
                if (w <= 0 || h <= 0) continue;
                if (abort) return;
                blit(backbuffer, x, y, x, y, w + x, h + y);
            }
        }

        /** This is how subclasses signal a 'shallow dirty', indicating that although the backbuffer is valid, the screen is not */
        public final void Dirty(int x, int y, int w, int h) {
            screenDirtyRegions.dirty(x, y, w, h);
            Scheduler.renderAll();
        }

        public void dirty(int x, int y, int w, int h) {
            screenDirtyRegions.dirty(x, y, w, h);
            super.dirty(x, y, w, h);
        }

        /** copies a region from the doublebuffer to this surface */
        public abstract void blit(PixelBuffer source, int sx, int sy, int dx, int dy, int dx2, int dy2);

    }

}
